@@ -1,4 +1,4 @@
-/* $Id: tap-provider.c 1266 2006-03-30 23:34:20Z benny $ */
+/* $Id: tap-provider.c 2033 2006-09-13 23:36:30Z benny $ */
 /*-
  * Copyright (c) 2006 Benedikt Meurer <benny@xfce.org>
  *
@@ -28,10 +28,11 @@
 
 #include <thunar-vfs/thunar-vfs.h>
 
+#include <thunar-archive-plugin/tap-backend.h>
 #include <thunar-archive-plugin/tap-provider.h>
 
-/* use access(2) with older GLib */
-#if GLIB_CHECK_VERSION(2,8,0)
+/* use g_access() on win32 */
+#if defined(G_OS_WIN32)
 #include <glib/gstdio.h>
 #else
 #define g_access(filename, mode) access((filename), (mode))
@@ -46,10 +47,20 @@ static void   tap_provider_finalize             (GObject                  *objec
 static GList *tap_provider_get_file_actions     (ThunarxMenuProvider      *menu_provider,
                                                  GtkWidget                *window,
                                                  GList                    *files);
-static void   tap_provider_execute              (TapProvider              *tap_provider,
+#if THUNARX_CHECK_VERSION(0,4,1)
+static GList *tap_provider_get_dnd_actions      (ThunarxMenuProvider      *menu_provider,
                                                  GtkWidget                *window,
-                                                 const gchar              *working_directory,
-                                                 const gchar              *command,
+                                                 ThunarxFileInfo          *folder,
+                                                 GList                    *files);
+#endif
+static void   tap_provider_execute              (TapProvider              *tap_provider,
+                                                 GPid                    (*action) (const gchar *folder,
+                                                                                    GList       *files,
+                                                                                    GtkWidget   *window,
+                                                                                    GError     **error),
+                                                 GtkWidget                *window,
+                                                 const gchar              *folder,
+                                                 GList                    *files,
                                                  const gchar              *error_message);
 static void   tap_provider_child_watch          (GPid                      pid,
                                                  gint                      status,
@@ -67,7 +78,12 @@ struct _TapProvider
 {
   GObject         __parent__;
 
+#if !GTK_CHECK_VERSION(2,9,0)
+  /* GTK+ 2.9.0 and above provide an icon-name property
+   * for GtkActions, so we don't need the icon factory.
+   */
   GtkIconFactory *icon_factory;
+#endif
 
   /* child watch support for the last spawn command,
    * which allows us to refresh the folder contents
@@ -80,7 +96,7 @@ struct _TapProvider
 
 
 
-static const char *TAP_MIME_TYPES[] = {
+static const gchar TAP_MIME_TYPES[][34] = {
   "application/x-ar",
   "application/x-arj",
   "application/x-bzip",
@@ -108,6 +124,9 @@ static const char *TAP_MIME_TYPES[] = {
 };
 
 static GQuark tap_action_files_quark;
+#if THUNARX_CHECK_VERSION(0,4,1)
+static GQuark tap_action_folder_quark;
+#endif
 static GQuark tap_action_provider_quark;
 
 
@@ -124,8 +143,11 @@ tap_provider_class_init (TapProviderClass *klass)
 {
   GObjectClass *gobject_class;
 
-  /* determine the "tap-action-files" and "tap-action-provider" quarks */
+  /* determine the "tap-action-files", "tap-action-folder" and "tap-action-provider" quarks */
   tap_action_files_quark = g_quark_from_string ("tap-action-files");
+#if THUNARX_CHECK_VERSION(0,4,1)
+  tap_action_folder_quark = g_quark_from_string ("tap-action-folder");
+#endif
   tap_action_provider_quark = g_quark_from_string ("tap-action-provider");
 
   gobject_class = G_OBJECT_CLASS (klass);
@@ -138,6 +160,9 @@ static void
 tap_provider_menu_provider_init (ThunarxMenuProviderIface *iface)
 {
   iface->get_file_actions = tap_provider_get_file_actions;
+#if THUNARX_CHECK_VERSION(0,4,1)
+  iface->get_dnd_actions = tap_provider_get_dnd_actions;
+#endif
 }
 
 
@@ -145,6 +170,7 @@ tap_provider_menu_provider_init (ThunarxMenuProviderIface *iface)
 static void
 tap_provider_init (TapProvider *tap_provider)
 {
+#if !GTK_CHECK_VERSION(2,9,0)
   GtkIconSource *icon_source;
   GtkIconSet    *icon_set;
 
@@ -152,18 +178,33 @@ tap_provider_init (TapProvider *tap_provider)
   tap_provider->icon_factory = gtk_icon_factory_new ();
   gtk_icon_factory_add_default (tap_provider->icon_factory);
 
-  /* add the "thunar-archive-plugin" stock icon */
+  /* add the "tap-create" stock icon */
   icon_set = gtk_icon_set_new ();
   icon_source = gtk_icon_source_new ();
-  gtk_icon_source_set_icon_name (icon_source, "file-roller");
+  gtk_icon_source_set_icon_name (icon_source, "tap-create");
   gtk_icon_set_add_source (icon_set, icon_source);
-  gtk_icon_factory_add (tap_provider->icon_factory, "thunar-archive-plugin", icon_set);
+  gtk_icon_factory_add (tap_provider->icon_factory, "tap-create", icon_set);
   gtk_icon_source_free (icon_source);
   gtk_icon_set_unref (icon_set);
 
-  /* initialize the child watch support */
-  tap_provider->child_watch_path = NULL;
-  tap_provider->child_watch_id = -1;
+  /* add the "tap-extract" stock icon */
+  icon_set = gtk_icon_set_new ();
+  icon_source = gtk_icon_source_new ();
+  gtk_icon_source_set_icon_name (icon_source, "tap-extract");
+  gtk_icon_set_add_source (icon_set, icon_source);
+  gtk_icon_factory_add (tap_provider->icon_factory, "tap-extract", icon_set);
+  gtk_icon_source_free (icon_source);
+  gtk_icon_set_unref (icon_set);
+
+  /* add the "tap-extract-to" stock icon */
+  icon_set = gtk_icon_set_new ();
+  icon_source = gtk_icon_source_new ();
+  gtk_icon_source_set_icon_name (icon_source, "tap-extract-to");
+  gtk_icon_set_add_source (icon_set, icon_source);
+  gtk_icon_factory_add (tap_provider->icon_factory, "tap-extract-to", icon_set);
+  gtk_icon_source_free (icon_source);
+  gtk_icon_set_unref (icon_set);
+#endif /* !GTK_CHECK_VERSION(2,9,0) */
 }
 
 
@@ -175,7 +216,7 @@ tap_provider_finalize (GObject *object)
   GSource     *source;
 
   /* give up maintaince of any pending child watch */
-  if (G_UNLIKELY (tap_provider->child_watch_id >= 0))
+  if (G_UNLIKELY (tap_provider->child_watch_id != 0))
     {
       /* reset the callback function to g_spawn_close_pid() so the plugin can be
        * safely unloaded and the child will still not become a zombie afterwards.
@@ -185,9 +226,11 @@ tap_provider_finalize (GObject *object)
       g_source_set_callback (source, (GSourceFunc) g_spawn_close_pid, NULL, NULL);
     }
   
+#if !GTK_CHECK_VERSION(2,9,0)
   /* release our icon factory */
   gtk_icon_factory_remove_default (tap_provider->icon_factory);
   g_object_unref (G_OBJECT (tap_provider->icon_factory));
+#endif
 
   (*G_OBJECT_CLASS (tap_provider_parent_class)->finalize) (object);
 }
@@ -239,46 +282,17 @@ tap_is_parent_writable (ThunarxFileInfo *file_info)
 
 
 
-static gchar*
-tap_files_to_string (GList *files)
-{
-  GString *string;
-  gchar   *filename;
-  gchar   *uri;
-  GList   *lp;
-
-  string = g_string_new ("");
-  for (lp = files; lp != NULL; lp = lp->next)
-    {
-      /* determine the filename of the file */
-      uri = thunarx_file_info_get_uri (lp->data);
-      filename = g_filename_from_uri (uri, NULL, NULL);
-      g_free (uri);
-
-      /* check if we have a valid filename */
-      if (G_LIKELY (filename != NULL))
-        {
-          /* append the filename to the string */
-          g_string_append_printf (string, " \"%s\"", filename);
-          g_free (filename);
-        }
-    }
-
-  return g_string_free (string, FALSE);
-}
-
-
-
 static void
 tap_extract_here (GtkAction *action,
                   GtkWidget *window)
 {
-  TapProvider *tap_provider;
-  GList       *files;
-  gchar       *files_string;
-  gchar       *command;
-  gchar       *dirname;
-  gchar       *uri;
+#if THUNARX_CHECK_VERSION(0,4,1)
+  ThunarxFileInfo *folder;
+#endif
+  TapProvider     *tap_provider;
+  GList           *files;
+  gchar           *dirname;
+  gchar           *uri;
 
   /* determine the files associated with the action */
   files = g_object_get_qdata (G_OBJECT (action), tap_action_files_quark);
@@ -290,32 +304,40 @@ tap_extract_here (GtkAction *action,
   if (G_UNLIKELY (tap_provider == NULL))
     return;
 
-  /* determine the parent URI of the first selected file */
-  uri = thunarx_file_info_get_parent_uri (files->data);
-  if (G_UNLIKELY (uri == NULL))
-    return;
+#if THUNARX_CHECK_VERSION(0,4,1)
+  /* check if a folder was supplied (for the Drag'n'Drop action) */
+  folder = g_object_get_qdata (G_OBJECT (action), tap_action_folder_quark);
+  if (G_UNLIKELY (folder != NULL))
+    {
+      /* determine the URI of the supplied folder */
+      uri = thunarx_file_info_get_uri (folder);
+    }
+  else
+#endif
+    {
+      /* determine the parent URI of the first selected file */
+      uri = thunarx_file_info_get_parent_uri (files->data);
+    }
 
-  /* determine the directory of the first selected file */
-  dirname = g_filename_from_uri (uri, NULL, NULL);
-  g_free (uri);
+  /* verify that we have an URI */
+  if (G_LIKELY (uri != NULL))
+    {
+      /* determine the directory of the first selected file */
+      dirname = g_filename_from_uri (uri, NULL, NULL);
 
-  /* verify that we were able to determine a local path */
-  if (G_UNLIKELY (dirname == NULL))
-    return;
+      /* verify that we were able to determine a local path */
+      if (G_LIKELY (dirname != NULL))
+        {
+          /* execute the action */
+          tap_provider_execute (tap_provider, tap_backend_extract_here, window, dirname, files, _("Failed to extract files"));
 
-  /* generate the files list string */
-  files_string = tap_files_to_string (files);
+          /* release the dirname */
+          g_free (dirname);
+        }
 
-  /* generate the command line */
-  command = g_strdup_printf ("file-roller --extract-to=\"%s\" --force %s", dirname, files_string);
-
-  /* try to run the command */
-  tap_provider_execute (tap_provider, window, dirname, command, _("Failed to extract files"));
-
-  /* cleanup */
-  g_free (files_string);
-  g_free (dirname);
-  g_free (command);
+      /* release the URI */
+      g_free (uri);
+    }
 }
 
 
@@ -327,8 +349,6 @@ tap_extract_to (GtkAction *action,
   TapProvider *tap_provider;
   const gchar *default_dir;
   GList       *files;
-  gchar       *files_string;
-  gchar       *command;
 
   /* determine the files associated with the action */
   files = g_object_get_qdata (G_OBJECT (action), tap_action_files_quark);
@@ -340,9 +360,6 @@ tap_extract_to (GtkAction *action,
   if (G_UNLIKELY (tap_provider == NULL))
     return;
 
-  /* generate the files list string */
-  files_string = tap_files_to_string (files);
-
   /* if $GTK_DEFAULT_FILECHOOSER_DIR is set, we use that as default
    * folder (i.e. Ubuntu), otherwise we just use $HOME.
    */
@@ -350,15 +367,8 @@ tap_extract_to (GtkAction *action,
   if (G_LIKELY (default_dir == NULL))
     default_dir = g_get_home_dir ();
 
-  /* generate the command line */
-  command = g_strdup_printf ("file-roller --default-dir=\"%s\" --extract %s", default_dir, files_string);
-
-  /* try to run the command */
-  tap_provider_execute (tap_provider, window, default_dir, command, _("Failed to extract files"));
-
-  /* cleanup */
-  g_free (files_string);
-  g_free (command);
+  /* execute the action */
+  tap_provider_execute (tap_provider, tap_backend_extract_to, window, default_dir, files, _("Failed to extract files"));
 }
 
 
@@ -369,8 +379,6 @@ tap_create_archive (GtkAction *action,
 {
   TapProvider *tap_provider;
   GList       *files;
-  gchar       *files_string;
-  gchar       *command;
   gchar       *dirname;
   gchar       *uri;
 
@@ -397,19 +405,11 @@ tap_create_archive (GtkAction *action,
   if (G_UNLIKELY (dirname == NULL))
     return;
 
-  /* generate the files list string */
-  files_string = tap_files_to_string (files);
-
-  /* generate the command line */
-  command = g_strdup_printf ("file-roller --default-dir=\"%s\" --add %s", dirname, files_string);
-
-  /* try to run the command */
-  tap_provider_execute (tap_provider, window, dirname, command, _("Failed to create archive"));
+  /* execute the action */
+  tap_provider_execute (tap_provider, tap_backend_create_archive, window, dirname, files, _("Failed to create archive"));
 
   /* cleanup */
-  g_free (files_string);
   g_free (dirname);
-  g_free (command);
 }
 
 
@@ -419,22 +419,29 @@ tap_provider_get_file_actions (ThunarxMenuProvider *menu_provider,
                                GtkWidget           *window,
                                GList               *files)
 {
-  TapProvider *tap_provider = TAP_PROVIDER (menu_provider);
-  GtkAction   *action;
-  GClosure    *closure;
-  gboolean     all_archives = TRUE;
-  gboolean     can_write = TRUE;
-  GList       *actions = NULL;
-  GList       *lp;
-  gint         n_files = 0;
-
-  /* verify that atleast one file is given */
-  if (G_UNLIKELY (files == NULL))
-    return NULL;
+  ThunarVfsPathScheme scheme;
+  ThunarVfsInfo      *info;
+  TapProvider        *tap_provider = TAP_PROVIDER (menu_provider);
+  GtkAction          *action;
+  GClosure           *closure;
+  gboolean            all_archives = TRUE;
+  gboolean            can_write = TRUE;
+  GList              *actions = NULL;
+  GList              *lp;
+  gint                n_files = 0;
 
   /* check all supplied files */
   for (lp = files; lp != NULL; lp = lp->next, ++n_files)
     {
+      /* check if the file is a local file */
+      info = thunarx_file_info_get_vfs_info (lp->data);
+      scheme = thunar_vfs_path_get_scheme (info->path);
+      thunar_vfs_info_unref (info);
+
+      /* unable to handle non-local files */
+      if (G_UNLIKELY (scheme != THUNAR_VFS_PATH_SCHEME_FILE))
+        return NULL;
+
       /* check if this file is a supported archive */
       if (all_archives && !tap_is_archive (lp->data))
         all_archives = FALSE;
@@ -451,12 +458,19 @@ tap_provider_get_file_actions (ThunarxMenuProvider *menu_provider,
       if (G_LIKELY (can_write))
         {
           /* append the "Extract Here" action */
-          action = gtk_action_new ("Tap::extract-here", _("Extract _Here"),
-                                   dngettext (GETTEXT_PACKAGE,
-                                              "Extract the selected archive in the current folder",
-                                              "Extract the selected archives in the current folder",
-                                              n_files),
-                                   "thunar-archive-plugin");
+          action = g_object_new (GTK_TYPE_ACTION,
+                                 "name", "Tap::extract-here",
+                                 "label", _("Extract _Here"),
+#if !GTK_CHECK_VERSION(2,9,0)
+                                 "stock-id", "tap-extract",
+#else
+                                 "icon-name", "tap-extract",
+#endif
+                                 "tooltip", dngettext (GETTEXT_PACKAGE,
+                                                       "Extract the selected archive in the current folder",
+                                                       "Extract the selected archives in the current folder",
+                                                       n_files),
+                                 NULL);
           g_object_set_qdata_full (G_OBJECT (action), tap_action_files_quark,
                                    thunarx_file_info_list_copy (files),
                                    (GDestroyNotify) thunarx_file_info_list_free);
@@ -469,12 +483,19 @@ tap_provider_get_file_actions (ThunarxMenuProvider *menu_provider,
         }
 
       /* append the "Extract To..." action */
-      action = gtk_action_new ("Tap::extract-to", _("_Extract To..."),
-                               dngettext (GETTEXT_PACKAGE,
-                                          "Extract the selected archive",
-                                          "Extract the selected archives",
-                                          n_files),
-                               "thunar-archive-plugin");
+      action = g_object_new (GTK_TYPE_ACTION,
+                             "label", _("_Extract To..."),
+                             "name", "Tap::extract-to",
+#if !GTK_CHECK_VERSION(2,9,0)
+                             "stock-id", "tap-extract-to",
+#else
+                             "icon-name", "tap-extract-to",
+#endif
+                             "tooltip", dngettext (GETTEXT_PACKAGE,
+                                                   "Extract the selected archive",
+                                                   "Extract the selected archives",
+                                                   n_files),
+                             NULL);
       g_object_set_qdata_full (G_OBJECT (action), tap_action_files_quark,
                                thunarx_file_info_list_copy (files),
                                (GDestroyNotify) thunarx_file_info_list_free);
@@ -490,12 +511,19 @@ tap_provider_get_file_actions (ThunarxMenuProvider *menu_provider,
   if (G_LIKELY (n_files > 1 || !all_archives))
     {
       /* append the "Create Archive..." action */
-      action = gtk_action_new ("Tap::create-archive", _("Cr_eate Archive..."),
-                               dngettext (GETTEXT_PACKAGE,
-                                          "Create an archive with the selected object",
-                                          "Create an archive with the selected objects",
-                                          n_files),
-                               "thunar-archive-plugin");
+      action = g_object_new (GTK_TYPE_ACTION,
+                             "label", _("Cr_eate Archive..."),
+                             "name", "Tap::create-archive",
+#if !GTK_CHECK_VERSION(2,9,0)
+                             "stock-id", "tap-create",
+#else
+                             "icon-name", "tap-create",
+#endif
+                             "tooltip", dngettext (GETTEXT_PACKAGE,
+                                                   "Create an archive with the selected object",
+                                                   "Create an archive with the selected objects",
+                                                   n_files),
+                             NULL);
       g_object_set_qdata_full (G_OBJECT (action), tap_action_files_quark,
                                thunarx_file_info_list_copy (files),
                                (GDestroyNotify) thunarx_file_info_list_free);
@@ -512,56 +540,120 @@ tap_provider_get_file_actions (ThunarxMenuProvider *menu_provider,
 
 
 
+#if THUNARX_CHECK_VERSION(0,4,1)
+static GList*
+tap_provider_get_dnd_actions (ThunarxMenuProvider *menu_provider,
+                              GtkWidget           *window,
+                              ThunarxFileInfo     *folder,
+                              GList               *files)
+{
+  ThunarVfsPathScheme scheme;
+  ThunarVfsInfo      *info;
+  TapProvider        *tap_provider = TAP_PROVIDER (menu_provider);
+  GtkAction          *action;
+  GClosure           *closure;
+  GList              *lp;
+  gint                n_files = 0;
+
+  /* check if the folder is a local folder */
+  info = thunarx_file_info_get_vfs_info (folder);
+  scheme = thunar_vfs_path_get_scheme (info->path);
+  thunar_vfs_info_unref (info);
+
+  /* unable to extract to non-local folders */
+  if (G_UNLIKELY (scheme != THUNAR_VFS_PATH_SCHEME_FILE))
+    return NULL;
+
+  /* check all supplied files */
+  for (lp = files; lp != NULL; lp = lp->next, ++n_files)
+    {
+      /* check if the file is a local file */
+      info = thunarx_file_info_get_vfs_info (lp->data);
+      scheme = thunar_vfs_path_get_scheme (info->path);
+      thunar_vfs_info_unref (info);
+
+      /* unable to handle non-local files */
+      if (G_UNLIKELY (scheme != THUNAR_VFS_PATH_SCHEME_FILE))
+        return NULL;
+
+      /* check if this file is a supported archive */
+      if (G_LIKELY (!tap_is_archive (lp->data)))
+        return NULL;
+    }
+
+  /* setup the "Extract here" action */
+  action = g_object_new (GTK_TYPE_ACTION,
+                         "name", "Tap::extract-here-dnd",
+                         /* TRANSLATORS: This is the label of the Drag'n'Drop "Extract here" action */
+                         "label", _("_Extract here"),
+#if !GTK_CHECK_VERSION(2,9,0)
+                         "stock-id", "tap-extract",
+#else
+                         "icon-name", "tap-extract",
+#endif
+                         "tooltip", dngettext (GETTEXT_PACKAGE,
+                                               "Extract the selected archive here",
+                                               "Extract the selected archives here",
+                                               n_files),
+                         NULL);
+  g_object_set_qdata_full (G_OBJECT (action), tap_action_files_quark,
+                           thunarx_file_info_list_copy (files),
+                           (GDestroyNotify) thunarx_file_info_list_free);
+  g_object_set_qdata_full (G_OBJECT (action), tap_action_provider_quark,
+                           g_object_ref (G_OBJECT (tap_provider)),
+                           (GDestroyNotify) g_object_unref);
+  g_object_set_qdata_full (G_OBJECT (action), tap_action_folder_quark,
+                           g_object_ref (G_OBJECT (folder)),
+                           (GDestroyNotify) g_object_unref);
+  closure = g_cclosure_new_object (G_CALLBACK (tap_extract_here), G_OBJECT (window));
+  g_signal_connect_closure (G_OBJECT (action), "activate", closure, TRUE);
+
+  /* return a list with only the "Extract here" action */
+  return g_list_prepend (NULL, action);
+}
+#endif
+
+
+
 static void
 tap_provider_execute (TapProvider *tap_provider,
+                      GPid       (*action) (const gchar *folder,
+                                            GList       *files,
+                                            GtkWidget   *window,
+                                            GError     **error),
                       GtkWidget   *window,
-                      const gchar *working_directory,
-                      const gchar *command_line,
+                      const gchar *folder,
+                      GList       *files,
                       const gchar *error_message)
 {
   GtkWidget *dialog;
-  GdkScreen *screen;
-  gboolean   succeed;
   GSource   *source;
   GError    *error = NULL;
-  gchar    **argv;
-  gint       pid;
+  GPid       pid;
 
-  /* determine the screen on which to run the command */
-  screen = gtk_widget_get_screen (window);
-  if (G_UNLIKELY (screen == NULL))
-    return;
-
-  /* try to parse the command line */
-  succeed = g_shell_parse_argv (command_line, NULL, &argv, &error);
-  if (G_LIKELY (succeed))
+  /* try to execute the action */
+  pid = (*action) (folder, files, window, &error);
+  if (G_LIKELY (pid >= 0))
     {
-      /* try to run the command */
-      succeed = gdk_spawn_on_screen (screen, working_directory, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH, NULL, NULL, &pid, &error);
-      if (G_LIKELY (succeed))
+      /* check if we already have a child watch */
+      if (G_UNLIKELY (tap_provider->child_watch_id != 0))
         {
-          /* check if we already have a child watch */
-          if (G_UNLIKELY (tap_provider->child_watch_id >= 0))
-            {
-              /* reset the callback function to g_spawn_close_pid() so the plugin can be
-               * safely unloaded and the child will still not become a zombie afterwards.
-               */
-              source = g_main_context_find_source_by_id (NULL, tap_provider->child_watch_id);
-              g_source_set_callback (source, (GSourceFunc) g_spawn_close_pid, NULL, NULL);
-            }
-
-          /* schedule the new child watch */
-          tap_provider->child_watch_id = g_child_watch_add_full (G_PRIORITY_LOW, pid, tap_provider_child_watch,
-                                                                 tap_provider, tap_provider_child_watch_destroy);
-
-          
-          /* remember the working directory for the child watch */
-          tap_provider->child_watch_path = g_strdup (working_directory);
+          /* reset the callback function to g_spawn_close_pid() so the plugin can be
+           * safely unloaded and the child will still not become a zombie afterwards.
+           */
+          source = g_main_context_find_source_by_id (NULL, tap_provider->child_watch_id);
+          g_source_set_callback (source, (GSourceFunc) g_spawn_close_pid, NULL, NULL);
         }
-    }
 
-  /* check if we failed */
-  if (G_UNLIKELY (!succeed))
+      /* schedule the new child watch */
+      tap_provider->child_watch_id = g_child_watch_add_full (G_PRIORITY_LOW, pid, tap_provider_child_watch,
+                                                             tap_provider, tap_provider_child_watch_destroy);
+
+      
+      /* remember the working directory for the child watch */
+      tap_provider->child_watch_path = g_strdup (folder);
+    }
+  else if (error != NULL)
     {
       /* display an error dialog */
       dialog = gtk_message_dialog_new (GTK_WINDOW (window),
@@ -575,9 +667,6 @@ tap_provider_execute (TapProvider *tap_provider,
       gtk_widget_destroy (dialog);
       g_error_free (error);
     }
-
-  /* cleanup */
-  g_strfreev (argv);
 }
 
 
@@ -626,7 +715,7 @@ tap_provider_child_watch_destroy (gpointer user_data)
   /* reset child watch id and path */
   g_free (tap_provider->child_watch_path);
   tap_provider->child_watch_path = NULL;
-  tap_provider->child_watch_id = -1;
+  tap_provider->child_watch_id = 0;
 }
 
 
